#! /usr/bin/env python3

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
import os

class StanfordOnlineProducts(Dataset):
    def __init__(self, dataset_root, transform=None):
        self.dataset_folder = os.path.join(dataset_root, "Stanford_Online_Products")
        self.load_labels()
        self.transform = transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        path = self.img_paths[idx]
        img = Image.open(path).convert("RGB")
        label = self.labels[idx]
        if self.transform is not None:
            img = self.transform(img)
        output_dict = {"data": img, "label": label}
        return output_dict

    def load_labels(self):
        from pandas import read_csv
        info_files = {"train": "Ebay_train.txt", "test": "Ebay_test.txt"}
        self.img_paths = []
        self.labels = []
        global_idx = 0
        self.predefined_splits = {}
        for k, v in info_files.items():
            curr_file = read_csv(os.path.join(self.dataset_folder, v), delim_whitespace=True, header=0).values
            self.img_paths.extend([os.path.join(self.dataset_folder, x) for x in list(curr_file[:,3])])
            self.labels.extend(list(curr_file[:,1] - 1))
            self.predefined_splits[k] = np.arange(global_idx, global_idx + len(curr_file))
            global_idx += len(curr_file)
        self.labels = np.array(self.labels)



import sys
import copy
from ..utils import common_functions as c_f, split_manager
from pytorch_metric_learning import losses
import pytorch_metric_learning.utils.logging_presets as logging_presets
import pytorch_metric_learning.utils.common_functions as pml_cf
from torch.utils.tensorboard import SummaryWriter
import torch.nn
import torch
import os
import pathlib
import shutil
import logging
import numpy as np
from scipy import stats as scipy_stats
from collections import defaultdict
from .. import architectures


class BaseAPIParser:
    def __init__(self, args, pytorch_getter, global_db_path=None):
        pml_cf.NUMPY_RANDOM = np.random.RandomState()
        logging.info("NUMPY_RANDOM = %s"%pml_cf.NUMPY_RANDOM)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.args = args
        self.pytorch_getter = pytorch_getter
        self.experiment_folder = args.experiment_folder
        self.global_db_path = global_db_path

        _model_folder = os.path.join("%s", "%s", "saved_models")
        _csv_folder = os.path.join("%s", "%s", "saved_csvs")
        _tensorboard_folder = os.path.join("%s", "%s", "tensorboard_logs")
        self.sub_experiment_dirs = [
            _model_folder,
            _csv_folder,
            _tensorboard_folder
        ]

        self.trainer, self.tester_obj = None, None

    def run(self):
        if self.beginning_of_training():
            self.make_dir()
        self.set_split_manager()
        self.save_config_files()
        self.set_num_epochs_dict()
        self.make_sub_experiment_dirs()
        self.set_meta_record_keeper()
        if self.args.evaluate and self.args.meta_testing_method:
            self.meta_eval()
        else:
            try:
                self.run_for_each_split_scheme()
            except ValueError as value_exception:
                error_string = str(value_exception)
                if "NaN" in error_string:
                    logging.error(error_string)
                    mean, sem = 0, 0
                    self.delete_old_objects()
                    return mean, sem if hasattr(self, "meta_record_keeper") else mean
                else:
                    raise ValueError
            self.record_meta_logs()
        self.flush_tensorboard()
        self.delete_old_objects()
        if self.is_training():
            return self.return_val_accuracy_and_standard_error()

    def run_for_each_split_scheme(self):
        for self.curr_split_count, split_scheme_name in enumerate(self.split_manager.split_scheme_names):
            num_epochs = self.num_epochs[split_scheme_name]
            self.split_manager.set_curr_split_scheme(split_scheme_name)
            self.set_curr_folders()
            self.set_models_optimizers_losses()
            if self.args.evaluate:
                self.eval()
            elif self.should_train(num_epochs, split_scheme_name):
                self.train(num_epochs)
            self.update_meta_record_keeper(split_scheme_name)
            self.delete_old_objects()

    def delete_old_objects(self):
        for attr_name in ["models", "loss_funcs", "mining_funcs", "optimizers", "lr_schedulers", "gradient_clippers"]:
            try:
                delattr(self, attr_name)
                logging.info("Deleted self.%s"%attr_name)
            except AttributeError:
                pass
        try:
            del self.trainer.dataloader
            del self.trainer.dataloader_iter
            logging.info("Deleted trainer dataloader")
        except AttributeError:
            pass

    def is_training(self):
        return not self.args.evaluate

    def beginning_of_training(self):
        return (not self.args.resume_training) and (not self.args.evaluate)

    def make_dir(self):
        root = pathlib.Path(self.experiment_folder)
        non_empty_dirs = {str(p.parent) for p in root.rglob('*') if p.is_file()}
        non_empty_dirs.discard(self.experiment_folder)
        if len(non_empty_dirs) > 0:
            logging.info("Experiment folder already taken!")
            sys.exit()
        c_f.makedir_if_not_there(self.experiment_folder)

    def make_sub_experiment_dirs(self):
        for s in self.sub_experiment_dirs:
            for r in self.split_manager.split_scheme_names:
                c_f.makedir_if_not_there(s % (self.experiment_folder, r))

    def set_curr_folders(self):
        self.model_folder, self.csv_folder, self.tensorboard_folder = self.get_sub_experiment_dir_paths()[self.split_manager.curr_split_scheme_name]

    def get_sub_experiment_dir_paths(self):
        sub_experiment_dir_paths = {}
        for k in self.split_manager.split_scheme_names:
            sub_experiment_dir_paths[k] = [s % (self.experiment_folder, k) for s in self.sub_experiment_dirs]
        return sub_experiment_dir_paths

    def set_num_epochs_dict(self):
        if isinstance(self.args.num_epochs_train, int):
            self.num_epochs = {k: self.args.num_epochs_train for k in self.split_manager.split_scheme_names}
        else:
            self.num_epochs = self.args.num_epochs_train

    def save_config_files(self):
        self.latest_sub_experiment_epochs = c_f.latest_sub_experiment_epochs(self.get_sub_experiment_dir_paths())
        latest_epochs = list(self.latest_sub_experiment_epochs.values())
        if self.is_training():
            c_f.save_config_files(self.args.place_to_save_configs, self.args.dict_of_yamls, self.args.resume_training, self.args.reproduce_results, latest_epochs)
        delattr(self.args, "dict_of_yamls")
        delattr(self.args, "place_to_save_configs")

    def get_accuracy_calculator(self):
        return self.pytorch_getter.get("accuracy_calculator", yaml_dict=self.args.eval_accuracy_calculator)

    def set_optimizers(self):
        self.optimizers, self.lr_schedulers, self.gradient_clippers = {}, {}, {}
        for k, v in self.args.optimizers.items():
            basename = k.replace("_optimizer", '')
            param_source = None
            for possible_params in [self.models, self.loss_funcs]:
                if basename in possible_params:
                    param_source = possible_params[basename]
                    break
            o, s, g = self.pytorch_getter.get_optimizer(param_source, yaml_dict=v)
            logging.info("%s\n%s" % (k, o))
            if o is not None: self.optimizers[k] = o
            if s is not None: self.lr_schedulers = {"%s_%s"%(basename, k):v for k,v in s.items()}
            if g is not None: self.gradient_clippers[basename + "_grad_clipper"] = g

    def set_split_manager(self):
        chosen_dataset = self.pytorch_getter.get("dataset", yaml_dict=self.args.dataset, additional_params={"dataset_root":self.args.dataset_root})
        self.split_manager = split_manager.SplitManager(dataset=chosen_dataset, 
                                                        test_size=self.args.test_size,
                                                        test_start_idx=self.args.test_start_idx, 
                                                        num_training_partitions=self.args.num_training_partitions,
                                                        num_training_sets=self.args.num_training_sets,
                                                        special_split_scheme_name=self.args.special_split_scheme_name,
                                                        hierarchy_level=self.args.label_hierarchy_level)
        if not self.args.splits_to_eval: self.args.splits_to_eval = [x for x in self.split_manager.split_names if x!="test"]

    def get_transforms(self):
        try:
            trunk = self.models["trunk"]
            if isinstance(trunk, torch.nn.DataParallel):
                trunk = trunk.module
            model_transform_properties = {k:getattr(trunk, k) for k in ["mean", "std", "input_space", "input_range"]}
        except (KeyError, AttributeError):
            model_transform_properties = {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]}
        transforms = {"train": None, "eval": None}
        for k, v in self.args.transforms.items():
            transforms[k] = self.pytorch_getter.get_composed_img_transform(v, **model_transform_properties)
        return transforms

    def set_transforms(self):
        logging.info("Setting dataset so that transform can be set")
        self.split_manager.set_curr_split("train", is_training=True)
        transforms = self.get_transforms()
        self.split_manager.set_transforms(transforms["train"], transforms["eval"])

    def get_embedder_model(self, model_type, input_size=None, output_size=None):
        model, model_args = self.pytorch_getter.get("model", yaml_dict=model_type, return_uninitialized=True)
        model_args = copy.deepcopy(model_args)
        if model == architectures.misc_models.MLP:
            if input_size:
                model_args["layer_sizes"].insert(0, input_size)
            if output_size:
                model_args["layer_sizes"].append(output_size)
        model = model(**model_args)
        logging.info("EMBEDDER MODEL %s"%model)
        return model

    def get_trunk_model(self, model_type):
        model = self.pytorch_getter.get("model", yaml_dict=model_type)
        self.base_model_output_size = c_f.get_last_linear(model).in_features
        c_f.set_last_linear(model, architectures.misc_models.Identity())
        return model

    def get_mining_function(self, miner_type):
        if miner_type:
            miner, miner_params = self.pytorch_getter.get("miner", yaml_dict=miner_type, return_uninitialized=True)
            miner_params = copy.deepcopy(miner_params)
            if "loss" in miner_params: miner_params["loss"] = self.get_loss_function(miner_params["loss"])
            if "miner" in miner_params: miner_params["miner"] = self.get_mining_function(miner_params["miner"])
            return miner(**miner_params)
        return None

    def set_sampler(self):
        if self.args.sampler in [None, {}]:
            self.sampler = None
        else:
            self.sampler = self.pytorch_getter.get("sampler", yaml_dict=self.args.sampler, additional_params={"labels":self.split_manager.labels})

    def get_loss_function(self, loss_type):
        loss, loss_params = self.pytorch_getter.get("loss", yaml_dict=loss_type, return_uninitialized=True)
        loss_params = copy.deepcopy(loss_params)
        if loss == losses.MultipleLosses:
            loss_funcs = [self.get_loss_function({k:v}) for k,v in loss_params["losses"].items()]
            return loss(loss_funcs) 
        if loss == losses.CrossBatchMemory:
            if "loss" in loss_params: loss_params["loss"] = self.get_loss_function(loss_params["loss"])
            if "miner" in loss_params: loss_params["miner"] = self.get_mining_function(loss_params["miner"])
        if c_f.check_init_arguments(loss, "num_classes"):
            loss_params["num_classes"] = self.split_manager.get_num_labels()
            logging.info("Passing %d as num_classes to the loss function"%loss_params["num_classes"])
        if "regularizer" in loss_params:
            loss_params["regularizer"] = self.pytorch_getter.get("regularizer", yaml_dict=loss_params["regularizer"])
        if "num_class_per_param" in loss_params and loss_params["num_class_per_param"]:
            loss_params["num_class_per_param"] = self.split_manager.get_num_labels()
            logging.info("Passing %d as num_class_per_param to the loss function"%loss_params["num_class_per_param"])

        return loss(**loss_params)        

    def set_loss_function(self):
        self.loss_funcs = {}
        for k, v in self.args.loss_funcs.items():
            self.loss_funcs[k] = self.get_loss_function(v)

    def set_mining_function(self):
        self.mining_funcs = {}
        for k, v in self.args.mining_funcs.items():
            self.mining_funcs[k] = self.get_mining_function(v)

    def model_getter_dict(self):
        return {"trunk": lambda model_type: self.get_trunk_model(model_type),
                "embedder": lambda model_type: self.get_embedder_model(model_type, input_size=self.base_model_output_size)}

    def set_model(self):
        self.models = {}
        for k, v in self.model_getter_dict().items():
            self.models[k] = v(self.args.models[k])

    def load_model_for_eval(self, suffix):
        untrained = suffix == "-1"
        trunk_model = self.get_trunk_model(self.args.models["trunk"])
        if not untrained:
            embedder_model = self.get_embedder_model(self.args.models["embedder"], self.base_model_output_size)
            pml_cf.load_dict_of_models(
                {"trunk": trunk_model, "embedder": embedder_model},
                suffix,
                self.model_folder,
                self.device
            )
        else:
            embedder_model = architectures.misc_models.Identity()
        return torch.nn.DataParallel(trunk_model), torch.nn.DataParallel(embedder_model)

    def get_splits_exclusion_list(self, splits_to_eval):
        if set(splits_to_eval) in [{"train"}, {"val"}, {"train", "val"}]:
            return ["test"]
        return []

    def eval_assertions(self, dataset_dict):
        for k, v in dataset_dict.items():
            dataset = self.split_manager.curr_split_scheme[k]
            assert v is dataset
            assert v.dataset.transform is self.split_manager.eval_transform

    def eval_model(self, epoch, suffix, splits_to_eval=None, load_model=False, skip_eval_if_already_done=True):
        logging.info("Launching evaluation for model %s"%suffix)
        if load_model:
            trunk_model, embedder_model = self.load_model_for_eval(suffix=suffix)
        else:
            trunk_model, embedder_model = self.models["trunk"], self.models["embedder"]
        trunk_model, embedder_model = trunk_model.to(self.device), embedder_model.to(self.device)
        splits_to_exclude = self.get_splits_exclusion_list(splits_to_eval)
        dataset_dict = self.split_manager.get_dataset_dict(inclusion_list=splits_to_eval, exclusion_list=splits_to_exclude, is_training=False)
        self.eval_assertions(dataset_dict)
        return self.hooks.run_tester_separately(self.tester_obj, dataset_dict, epoch, trunk_model, embedder_model, 
                                        splits_to_eval=splits_to_eval, collate_fn=None, skip_eval_if_already_done=skip_eval_if_already_done)

    def flush_tensorboard(self):
        for keeper in ["record_keeper", "meta_record_keeper"]:
            k = getattr(self, keeper, None)
            if k:
                k.tensorboard_writer.flush()
                k.tensorboard_writer.close()

    def set_record_keeper(self):
        is_new_experiment = self.beginning_of_training() and self.curr_split_count == 0
        self.record_keeper, _, _ = logging_presets.get_record_keeper(self.csv_folder, self.tensorboard_folder, self.global_db_path, self.args.experiment_name, is_new_experiment)

    def set_meta_record_keeper(self):
        is_new_experiment = self.beginning_of_training()
        if len(self.split_manager.split_scheme_names) > 1:
            _, csv_folder, tensorboard_folder = [s % (self.experiment_folder, "meta_logs") for s in self.sub_experiment_dirs]
            self.meta_record_keeper, _, _ = logging_presets.get_record_keeper(csv_folder, tensorboard_folder,  self.global_db_path, self.args.experiment_name, is_new_experiment)
            self.meta_accuracies = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))

    def update_meta_record_keeper(self, split_scheme_name):
        if hasattr(self, "meta_accuracies"):
            for split in self.args.splits_to_eval:
                best_split_accuracies, _ = self.hooks.get_accuracies_of_best_epoch(self.tester_obj, split, ignore_epoch=(-1, 0))
                untrained_accuracies = self.hooks.get_accuracies_of_epoch(self.tester_obj, split, -1)
                for accuracies, is_trained in [(untrained_accuracies, 0), (best_split_accuracies, 1)]:
                    if len(accuracies) > 0:
                        accuracy_keys = [k for k in accuracies[0].keys() if any(acc in k for acc in pml_ca.METRICS)]
                        for k in accuracy_keys:
                            self.meta_accuracies[split][is_trained][k][split_scheme_name] = accuracies[0][k]

    def record_meta_logs(self):
        if hasattr(self, "meta_accuracies") and len(self.meta_accuracies) > 0:
            for split in self.args.splits_to_eval:
                group_name = self.get_eval_record_name_dict("meta")[split]
                len_of_existing_records = c_f.try_getting_db_count(self.meta_record_keeper, group_name)
                for i, is_trained in enumerate([0, 1]):
                    curr_dict = self.meta_accuracies[split][is_trained]
                    if len(curr_dict) > 0:
                        averages = {k: np.mean(list(v.values())) for k, v in curr_dict.items()}
                        standard_errors = {"SEM_%s"%k: scipy_stats.sem(list(v.values())) for k, v in curr_dict.items()}
                        global_iteration = len_of_existing_records+i+1
                        self.meta_record_keeper.update_records(averages, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)
                        self.meta_record_keeper.update_records(standard_errors, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)
                        self.meta_record_keeper.update_records({"is_trained": is_trained}, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)
                        self.meta_record_keeper.update_records({"timestamp": c_f.get_datetime()}, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)
            self.meta_record_keeper.save_records()

    def return_val_accuracy_and_standard_error(self):
        if hasattr(self, "meta_record_keeper"):
            group_name = self.get_eval_record_name_dict("meta")["val"]
            def get_average_best_and_sem(key):
                is_trained = 1
                sem_key = "SEM_%s"%key
                query = "SELECT {0}, {1} FROM {2} WHERE is_trained=? AND id=(SELECT MAX(id) FROM {2})".format(key, sem_key, group_name)
                return self.meta_record_keeper.query(query, values=(is_trained,), use_global_db=False), key, sem_key
            q, key, sem_key = self.hooks.try_primary_metric(self.tester_obj, get_average_best_and_sem)
            return q[0][key], q[0][sem_key]
        _, best_accuracy = self.hooks.get_best_epoch_and_accuracy(self.tester_obj, "val", ignore_epoch=(-1,0))
        return best_accuracy


    def maybe_load_models_and_records(self):
        return self.hooks.load_latest_saved_models(self.trainer, self.model_folder, self.device)

    def set_models_optimizers_losses(self):
        self.set_model()
        self.set_transforms()
        self.set_sampler()
        self.set_loss_function()
        self.set_mining_function()
        self.set_optimizers()
        self.set_record_keeper()
        self.hooks = logging_presets.HookContainer(self.record_keeper, primary_metric=self.args.eval_primary_metric, validation_split_name="val")
        self.tester_obj = self.pytorch_getter.get("tester", self.args.testing_method, self.get_tester_kwargs())
        self.trainer = self.pytorch_getter.get("trainer", self.args.training_method, self.get_trainer_kwargs())
        self.epoch = self.maybe_load_models_and_records()
        self.set_dataparallel()
        self.set_devices()

    def get_tester_kwargs(self):
        return {
            "reference_set": self.args.eval_reference_set,
            "normalize_embeddings": self.args.eval_normalize_embeddings,
            "use_trunk_output": self.args.eval_use_trunk_output,
            "batch_size": self.args.eval_batch_size,
            "data_device": self.device,
            "dataloader_num_workers": self.args.eval_dataloader_num_workers,
            "pca": self.args.eval_pca,
            "size_of_tsne": self.args.eval_size_of_tsne,
            "data_and_label_getter": lambda data: (data["data"], data["label"]),
            "label_hierarchy_level": self.args.label_hierarchy_level,
            "end_of_testing_hook": self.hooks.end_of_testing_hook,
            "accuracy_calculator": self.get_accuracy_calculator() 
        }

    def get_end_of_epoch_hook(self):
        logging.info("Creating end_of_epoch_hook kwargs")
        splits_to_exclude = self.get_splits_exclusion_list(self.args.splits_to_eval)
        dataset_dict = self.split_manager.get_dataset_dict(inclusion_list=self.args.splits_to_eval, exclusion_list=splits_to_exclude, is_training=False)
        helper_hook = self.hooks.end_of_epoch_hook(tester=self.tester_obj,
                                                    dataset_dict=dataset_dict,
                                                    model_folder=self.model_folder,
                                                    test_interval=self.args.save_interval,
                                                    patience=self.args.patience,
                                                    test_collate_fn=None)
        def end_of_epoch_hook(trainer):
            torch.cuda.empty_cache()
            for k in dataset_dict.keys(): self.split_manager.set_curr_split(k, is_training=False)
            self.eval_assertions(dataset_dict)
            continue_training = helper_hook(trainer)
            for k in dataset_dict.keys(): self.split_manager.set_curr_split(k, is_training=True)
            self.training_assertions(trainer)
            return continue_training

        return end_of_epoch_hook

    def get_trainer_kwargs(self):
        return {
            "models": self.models,
            "optimizers": self.optimizers,
            "batch_size": self.args.batch_size,
            "sampler": self.sampler,
            "collate_fn": None,
            "loss_funcs": self.loss_funcs,
            "mining_funcs": self.mining_funcs,
            "dataset": self.split_manager.dataset,
            "data_device": self.device,
            "iterations_per_epoch": self.args.iterations_per_epoch,
            "lr_schedulers": self.lr_schedulers,
            "gradient_clippers": self.gradient_clippers,
            "freeze_trunk_batchnorm": self.args.freeze_batchnorm,
            "label_hierarchy_level": self.args.label_hierarchy_level,
            "dataloader_num_workers": self.args.dataloader_num_workers,
            "loss_weights": getattr(self.args, "loss_weights", None),
            "data_and_label_getter": lambda data: (data["data"], data["label"]),
            "dataset_labels": self.split_manager.labels,
            "set_min_label_to_zero": True,
            "end_of_iteration_hook": self.hooks.end_of_iteration_hook,
            "end_of_epoch_hook": self.get_end_of_epoch_hook()
        }

    def set_dataparallel(self):
        for k, v in self.models.items():
            self.models[k] = torch.nn.DataParallel(v)

    def set_devices(self):
        for obj_dict in [self.models, self.loss_funcs, self.mining_funcs]:
            for v in obj_dict.values():
                v.to(self.device)
        for v in self.optimizers.values():
            c_f.move_optimizer_to_gpu(v, self.device)

    def should_train(self, num_epochs, split_scheme_name):
        best_epoch, _ = self.hooks.get_best_epoch_and_accuracy(self.tester_obj, "val", ignore_epoch=(-1,))
        return self.hooks.patience_remaining(self.epoch, best_epoch, self.args.patience) and self.latest_sub_experiment_epochs[split_scheme_name] < num_epochs

    def training_assertions(self, trainer):
        dataset = self.split_manager.curr_split_scheme["train"]
        assert trainer.dataset is dataset
        assert trainer.dataset.dataset.transform is self.split_manager.train_transform
        assert np.array_equal(trainer.dataset_labels, self.split_manager.original_dataset.labels[dataset.indices])

    def train(self, num_epochs):
        if self.args.check_untrained_accuracy:
            for epoch, name, load_model in [(-1, "-1", True), (0, "0", False)]:
                self.eval_model(epoch, name,  splits_to_eval=self.args.splits_to_eval, load_model=load_model, skip_eval_if_already_done=self.args.skip_eval_if_already_done)
                self.record_keeper.save_records()
        self.split_manager.set_curr_split("train", is_training=True)
        self.training_assertions(self.trainer)        
        self.trainer.train(self.epoch, num_epochs)
        self.epoch = self.trainer.epoch + 1

    def eval(self):
        best_epoch, _ = self.hooks.get_best_epoch_and_accuracy(self.tester_obj, "val", ignore_epoch=(-1,0))
        eval_dict = {"best": best_epoch}
        if self.args.check_untrained_accuracy: eval_dict["-1"] = -1
        for name, epoch in eval_dict.items():
            self.eval_model(epoch, name, splits_to_eval=self.args.splits_to_eval, load_model=True, skip_eval_if_already_done=self.args.skip_eval_if_already_done)
            self.record_keeper.save_records()

    def meta_ConcatenateEmbeddings(self, model_suffix): 
        list_of_trunks, list_of_embedders = [], []
        for split_scheme_name in self.split_manager.split_scheme_names:
            self.split_manager.set_curr_split_scheme(split_scheme_name)
            self.set_curr_folders()
            trunk_model, embedder_model = self.load_model_for_eval(suffix=model_suffix)
            list_of_trunks.append(trunk_model.module)
            list_of_embedders.append(embedder_model.module)
        embedder_input_sizes = [self.base_model_output_size] * len(list_of_trunks)
        if isinstance(embedder_input_sizes[0], list):
            embedder_input_sizes = [np.sum(x) for x in embedder_input_sizes]
        operation_before_concat = (lambda x: torch.nn.functional.normalize(x, p=2, dim=1)) if self.args.eval_normalize_embeddings else None
        trunk = torch.nn.DataParallel(architectures.misc_models.ListOfModels(list_of_trunks))
        embedder = torch.nn.DataParallel(architectures.misc_models.ListOfModels(list_of_embedders, embedder_input_sizes, operation_before_concat))
        return trunk, embedder

    def meta_eval(self):
        assert self.args.splits_to_eval == ["test"]
        meta_model_getter = getattr(self, "meta_"+self.args.meta_testing_method)
        self.models = {}
        self.record_keeper = self.meta_record_keeper
        self.hooks = logging_presets.HookContainer(self.record_keeper, record_group_name_prefix=meta_model_getter.__name__, primary_metric=self.args.eval_primary_metric)
        self.tester_obj = self.pytorch_getter.get("tester", 
                                                self.args.testing_method, 
                                                self.get_tester_kwargs())

        eval_dict = {"best": 1}
        if self.args.check_untrained_accuracy: eval_dict["-1"] = -1

        group_name = self.get_eval_record_name_dict("meta_ConcatenateEmbeddings")["test"]
        len_of_existing_records = c_f.try_getting_db_count(self.meta_record_keeper, group_name)

        for name, i in eval_dict.items():
            self.models["trunk"], self.models["embedder"] = meta_model_getter(name)
            self.set_transforms()
            did_not_skip = self.eval_model(i, name, splits_to_eval=self.args.splits_to_eval, load_model=False, skip_eval_if_already_done=self.args.skip_meta_eval_if_already_done)
            if did_not_skip:
                is_trained = int(i==1)
                global_iteration = len_of_existing_records + is_trained + 1
                self.record_keeper.update_records({"is_trained": int(i==1)}, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)
                self.record_keeper.update_records({"timestamp": c_f.get_datetime()}, global_iteration=global_iteration, input_group_name_for_non_objects=group_name)

                for irrelevant_key in ["epoch", "best_epoch", "best_accuracy"]:
                    self.record_keeper.record_writer.records.pop(irrelevant_key, None)
                self.record_keeper.save_records()


    def get_eval_record_name_dict(self, eval_type="non_meta", return_all=False):
        prefix = self.hooks.record_group_name_prefix 
        self.hooks.record_group_name_prefix = "" #temporary
        non_meta = {k:self.hooks.record_group_name(self.tester_obj, k) for k in ["train", "val", "test"]}
        meta = {k:"meta_"+v for k,v in non_meta.items()}
        meta_concatenated = {k:"meta_ConcatenateEmbeddings_"+v for k,v in non_meta.items()}
        self.hooks.record_group_name_prefix = prefix

        name_dict = {"non_meta": non_meta,
                    "meta": meta,
                    "meta_ConcatenateEmbeddings": meta_concatenated}

        if return_all:
            return name_dict
        return name_dict[eval_type]







#! /usr/bin/env python3

import numpy as np
from torch.utils.data import Dataset
from torchvision import datasets
import os

class CUB200(Dataset):
    def __init__(self, dataset_root, transform=None):
        img_folder = os.path.join(dataset_root, "cub2011", "CUB_200_2011", "images")
        self.dataset = datasets.ImageFolder(img_folder)
        self.labels = np.array([b for (a, b) in self.dataset.imgs])
        self.transform = transform

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        img, label = self.dataset[idx]
        if self.transform is not None:
            img = self.transform(img)
        output_dict = {"data": img, "label": label}
        return output_dict



from .base_api_parser import BaseAPIParser
APIMetricLossOnly = BaseAPIParser
from .api_cascaded_embeddings import APICascadedEmbeddings
from .api_deep_adversarial_metric_learning import APIDeepAdversarialMetricLearning
from .api_train_with_classifier import APITrainWithClassifier
from .api_unsupervised_embeddings_using_augmentations import APIUnsupervisedEmbeddingsUsingAugmentations



#! /usr/bin/env python3

from collections import OrderedDict

import numpy as np
import torch
from . import dataset_utils as d_u
import logging
import itertools

class SplitManager:
    def __init__(
        self,
        dataset=None,
        train_transform=None,
        eval_transform=None,
        test_size=None,
        test_start_idx=None,
        num_training_partitions=2,
        num_training_sets=1,
        special_split_scheme_name=None,
        hierarchy_level=0
    ):
        self.original_dataset = dataset
        self.train_transform = train_transform
        self.eval_transform = eval_transform
        self.test_size = test_size
        self.test_start_idx = test_start_idx
        self.num_training_partitions = num_training_partitions
        self.num_training_sets = num_training_sets
        self.special_split_scheme_name = special_split_scheme_name
        self.hierarchy_level = hierarchy_level
        self.is_training = True
        self.create_split_schemes()

    def assert_splits_are_class_disjoint(self):
        for (split_scheme_name, split_scheme) in self.split_schemes.items():
            labels = []
            for split, dataset in split_scheme.items():
                labels.append(set(d_u.get_labels_by_hierarchy(self.original_dataset.labels[dataset.indices], self.hierarchy_level)))
            for (x,y) in itertools.combinations(labels, 2):
                assert x.isdisjoint(y)

    def assert_same_test_set_across_schemes(self):
        test_key = "val" if self.special_split_scheme_name == "old_approach" else "test"
        prev_indices = None
        for (split_scheme_name, split_scheme) in self.split_schemes.items():
            curr_indices = np.array(split_scheme[test_key].indices)
            if prev_indices is not None:
                assert np.array_equal(curr_indices, prev_indices)
            prev_indices = curr_indices


    def create_split_schemes(self):
        """
        Creates a dictionary where each key is the name of a split_scheme.
        Each value is a dictionary with "train", "val", "test" keys, corresponding
        to subsets of self.original_dataset
        """
        self.split_schemes = OrderedDict()
        if self.special_split_scheme_name:
            self.split_schemes[self.special_split_scheme_name] = d_u.create_one_split_scheme(self.original_dataset, 
                                                                                            scheme_name=self.special_split_scheme_name)
        else:
            for partition in range(self.num_training_sets):
                name = d_u.get_base_split_name(self.test_size, self.test_start_idx, self.num_training_partitions, partition=partition)
                self.split_schemes[name] = d_u.create_one_split_scheme(self.original_dataset, 
                                                                        partition=partition,
                                                                        num_training_partitions=self.num_training_partitions,
                                                                        test_size=self.test_size, 
                                                                        test_start_idx=self.test_start_idx,
                                                                        hierarchy_level=self.hierarchy_level)
        if self.special_split_scheme_name != "predefined": 
            self.assert_splits_are_class_disjoint()
            self.assert_same_test_set_across_schemes()
        self.split_scheme_names = list(self.split_schemes.keys())
        self.split_names = list(self.split_schemes[self.split_scheme_names[0]].keys())

    def set_curr_split_scheme(self, split_scheme_name):
        self.curr_split_scheme_name = split_scheme_name
        self.curr_split_scheme = self.split_schemes[self.curr_split_scheme_name]

    def set_curr_split(self, split_name, is_training, log_split_details=False):
        """
        Sets self.dataset and self.labels to a specific split within the current
        split_scheme.
        """
        self.curr_split_name = split_name
        self.is_training = is_training
        transform = self.train_transform if self.is_training else self.eval_transform
        self.dataset = self.curr_split_scheme[self.curr_split_name]
        self.set_dataset_transform(transform)
        if self.is_training:
            self.labels = self.original_dataset.labels[self.dataset.indices]
        if log_split_details:
            logging.info("SPLIT: %s / %s / length %d" % (self.curr_split_scheme_name, self.curr_split_name, len(self.dataset)))

    def set_transforms(self, train_transform, eval_transform):
        self.train_transform = train_transform
        self.eval_transform = eval_transform
        if self.is_training:
            self.set_dataset_transform(self.train_transform)
        else:
            self.set_dataset_transform(self.eval_transform)

    def set_dataset_transform(self, transform):
        self.dataset.dataset.transform = transform

    def get_num_labels(self):
        self.set_curr_split("train", True)
        L = np.array(self.labels)
        if L.ndim == 2:
            L = L[:, self.hierarchy_level]
        return len(set(L))

    def get_dataset_dict(self, inclusion_list=None, exclusion_list=None, is_training=False):
        logging.info("COLLECTING DATASETS FOR EVAL")
        dataset_dict = {}
        inclusion_list = list(self.curr_split_scheme.keys()) if inclusion_list is None else inclusion_list
        exclusion_list = [] if exclusion_list is None else exclusion_list
        allowed_list = [x for x in inclusion_list if x not in exclusion_list]
        for split_name, _ in self.curr_split_scheme.items():
            if split_name in allowed_list:
                self.set_curr_split(split_name, is_training, log_split_details=True)
                dataset_dict[split_name] = self.dataset
        return dataset_dict



#! /usr/bin/env python3
import logging
logging.info("Importing packages in single_experiment_runner")
from easy_module_attribute_getter import YamlReader, PytorchGetter
from ..utils import common_functions as c_f, dataset_utils as d_u
import argparse
import glob
import os
logging.info("Done importing packages in single_experiment_runner")


class SingleExperimentRunner:
    def __init__(self, 
                dataset_root="datasets", 
                root_experiment_folder="experiments", 
                pytorch_home=None, 
                root_config_folder=None, 
                config_foldernames=None,
                global_db_path=None):
        self.dataset_root = dataset_root
        self.root_experiment_folder = root_experiment_folder
        self.global_db_path = global_db_path
        if pytorch_home is not None:
            os.environ["TORCH_HOME"] = pytorch_home
        if root_config_folder is not None:
            self.root_config_folder = root_config_folder
        else:
            self.root_config_folder = os.path.join(os.path.dirname(__file__), "../configs")
        self.init_pytorch_getter()
        self.set_config_foldernames(config_foldernames)
        self.set_YR()
        

    def register(self, module_type, module):
        self.pytorch_getter.register(module_type, module)

    def init_pytorch_getter(self):
        from pytorch_metric_learning import trainers, losses, miners, regularizers, samplers, testers, utils
        from .. import architectures
        from .. import datasets
        from .. import api_parsers
        self.pytorch_getter = PytorchGetter(use_pretrainedmodels_package=True)
        self.pytorch_getter.register('model', architectures.misc_models)
        self.pytorch_getter.register('loss', losses)
        self.pytorch_getter.register('miner', miners)
        self.pytorch_getter.register('regularizer', regularizers)
        self.pytorch_getter.register('sampler', samplers)
        self.pytorch_getter.register('trainer', trainers)
        self.pytorch_getter.register('tester', testers)
        self.pytorch_getter.register('dataset', datasets)
        self.pytorch_getter.register('api_parser', api_parsers)
        self.pytorch_getter.register('accuracy_calculator', utils.AccuracyCalculator)

    def set_YR(self):
        self.YR = self.setup_yaml_reader()


    def set_config_foldernames(self, config_foldernames=None):
        if config_foldernames:
            self.config_foldernames = config_foldernames
        else:
            self.config_foldernames = ["config_general", "config_models", "config_optimizers",
                                        "config_loss_and_miners", "config_transforms", "config_eval"]


    def run(self):
        if self.YR.args.reproduce_results:
            self.reproduce_results(self.YR)
        else:
            self.run_new_experiment(self.YR)

    def start_experiment(self, args):
        api_parser_kwargs = {"args": args, "pytorch_getter": self.pytorch_getter, "global_db_path": self.global_db_path}
        api_parser = self.pytorch_getter.get('api_parser', class_name="API"+args.training_method, params=api_parser_kwargs)
        run_output = api_parser.run()
        self.eval_record_group_dicts = api_parser.get_eval_record_name_dict(return_all=True)
        del api_parser.tester_obj
        del api_parser.trainer
        del api_parser
        return run_output

    def run_new_experiment(self, YR):
        args, _, args.dict_of_yamls = YR.load_yamls(**self.determine_where_to_get_yamls(YR.args), max_merge_depth=float('inf'))
        return self.start_experiment(args)

    def reproduce_results(self, YR):
        configs_folder = os.path.join(YR.args.reproduce_results, 'configs')
        default_configs, experiment_configs = [], []
        for k in self.config_foldernames:
            default_configs.append(os.path.join(self.root_config_folder, k, "default.yaml")) #append default config file
        for k in self.config_foldernames:
            experiment_configs.append(os.path.join(configs_folder, "%s.yaml"%k)) #append experiment config file
        args, _, args.dict_of_yamls = YR.load_yamls(config_paths=default_configs+experiment_configs, max_merge_depth=0)

        # check if there were config diffs if training was resumed
        if YR.args.special_split_scheme_name:
            num_training_sets = 1
            base_split_name = YR.args.special_split_scheme_name
        else:
            num_training_sets = YR.args.num_training_sets
            base_split_name = d_u.get_base_split_name(YR.args.test_size, YR.args.test_start_idx, YR.args.num_training_partitions)
        resume_training_dict = c_f.get_all_resume_training_config_diffs(configs_folder, base_split_name, num_training_sets)
        if len(resume_training_dict) > 0:
            for sub_folder, num_epochs_dict in resume_training_dict.items():
                # train until the next config diff was made
                args.num_epochs_train = num_epochs_dict
                self.start_experiment(args)
                # start with a fresh set of args
                YR = self.setup_yaml_reader()
                # load the default configs, the experiment specific configs, plus the config diffs 
                for k in glob.glob(os.path.join(sub_folder, "*")):
                    experiment_configs.append(k)
                args, _, args.dict_of_yamls = YR.load_yamls(config_paths=default_configs+experiment_configs, max_merge_depth=0)
                # remove default configs from dict_of_yamls to avoid saving these as config diffs
                for d in default_configs:
                    args.dict_of_yamls.pop(d, None)
                args.resume_training = True
        return self.start_experiment(args)



    def setup_argparser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("--experiment_name", type=str, required=True)
        parser.add_argument("--resume_training", action="store_true")
        parser.add_argument("--evaluate", action="store_true")
        parser.add_argument("--splits_to_eval", nargs="+", type=str, default=["val"])
        parser.add_argument("--reproduce_results", type=str, default=None)
        for c in self.config_foldernames:
            parser.add_argument("--%s" % c, nargs="+", type=str, required=False, default=["default"])
        return parser


    def setup_yaml_reader(self):
        argparser = self.setup_argparser()
        YR = YamlReader(argparser=argparser)
        YR.args.dataset_root = self.dataset_root
        YR.args.experiment_folder = os.path.join(self.root_experiment_folder, YR.args.experiment_name)
        YR.args.place_to_save_configs = os.path.join(YR.args.experiment_folder, "configs")
        return YR


    def determine_where_to_get_yamls(self, args):
        if args.resume_training or args.evaluate:
            config_paths = [os.path.join(args.place_to_save_configs,'%s.yaml'%v) for v in self.config_foldernames]
        else:
            config_paths = []
            for subfolder in self.config_foldernames:
                for curr_yaml in getattr(args, subfolder):
                    config_paths.append(os.path.join(self.root_config_folder, subfolder, "%s.yaml"%curr_yaml))
        return {"config_paths": config_paths}


#! /usr/bin/env python3

from collections import OrderedDict, defaultdict

import numpy as np
import torch.utils.data


def get_labels_by_hierarchy(labels, hierarchy_level):
    if labels.ndim == 2:
        labels = labels[:, hierarchy_level]
    return labels

def get_base_split_name(test_size, test_start_idx, num_training_partitions, partition=''):
    test_size = int(test_size*100)
    test_start_idx = int(test_start_idx*100)
    return 'Test%02d_%02d_Partitions%d_%s'%(test_size, test_start_idx, num_training_partitions, partition)


def create_subset(dataset, idx_to_keep):
    """
    Args:
        dataset: type torch.utils.data.Dataset, the dataset to return a subset of
        labels: type torch.tensor, the labels for every element in the dataset
        idx_to_keep: type sequence, the indices of the elements that will form
                     the subset
    Returns:
        subset_dataset: type torch.utils.data.Dataset
        subset_labels: type torch.tensor
    """
    assert max(idx_to_keep) < len(dataset)
    subset_dataset = torch.utils.data.Subset(dataset, idx_to_keep)
    return subset_dataset


def create_label_based_subset(dataset, labels, class_rule):
    """
    This behaves like create_subset, except for the third argument.
    class_rule is a function that takes in a label (an integer) and outputs
    True if that class should be included in the subset dataset.
    """
    idx_to_keep = [i for i, label in enumerate(labels) if class_rule(label)]
    return create_subset(dataset, idx_to_keep)


def numeric_class_rule(a, b, exclusion_rule=None):
    """
    Args:
        rule_params: a list, the purpose of which is determined by the mode specified
    Returns:
        a function that takes a label as input, and returns True if that label
                    should be included in a subset dataset.
    """
    exclusion_rule = (lambda label: True) if exclusion_rule is None else exclusion_rule

    if a < b:
        range_rule = lambda label: (a <= label <= b)
    elif a > b:
        range_rule = lambda label: (label >= a or label <= b)
    return lambda label: (range_rule(label) and exclusion_rule(label))


def get_wrapped_range(start_idx, length_of_range, length_of_list):
    s = start_idx % length_of_list
    e = (start_idx + length_of_range - 1) % length_of_list
    return s, e

def get_single_class_rule(start_idx, split_length, sorted_label_set, exclusion_rule=None):
    s, e = get_wrapped_range(start_idx, split_length, len(sorted_label_set))
    s = sorted_label_set[s]
    e = sorted_label_set[e]
    return numeric_class_rule(s, e, exclusion_rule)

def get_class_rules(start_idx, split_lengths, sorted_label_set, exclusion_rule=None):
    class_rules = {}
    for k, v in split_lengths.items():
        class_rules[k] = get_single_class_rule(start_idx, v, sorted_label_set, exclusion_rule)
        start_idx += v
    return class_rules

def split_lengths_from_ratios(class_ratios, num_labels):
    split_lengths = {k: int(num_labels * v) for k, v in class_ratios.items()}
    split_lengths["train"] += num_labels - sum(v for k, v in split_lengths.items())
    assert sum(v for _, v in split_lengths.items()) == num_labels
    return split_lengths

def create_one_split_scheme(dataset, scheme_name=None, partition=None, num_training_partitions=None, test_size=None, test_start_idx=None, hierarchy_level=0):
    """
    Args:
        dataset: type torch.utils.data.Dataset, the dataset to return a subset of
        split_scheme_name: type string, the name of the split scheme to be returned
    Returns:
        A dictionary where each key is a split name (i.e. "train", "val", "test"),
                and the value is a tuple: (subset_dataset, subset_labels)
    """
    traintest_dict = OrderedDict()
    if scheme_name == "predefined":
        for k, v in dataset.predefined_splits.items():
            traintest_dict[k] = torch.utils.data.Subset(dataset, v)
    else:
        labels = get_labels_by_hierarchy(dataset.labels, hierarchy_level)
        sorted_label_set = sorted(list(set(labels)))
        num_labels = len(sorted_label_set)

        if scheme_name == "old_approach":
            split_lengths = split_lengths_from_ratios({"train": 0.5, "val": 0.5}, num_labels)
            for k, class_rule in get_class_rules(0, split_lengths, sorted_label_set).items():
                traintest_dict[k] = create_label_based_subset(dataset, labels, class_rule)
        else:
            val_ratio = (1./num_training_partitions)*(1-test_size)
            train_ratio = (1. - val_ratio)*(1-test_size)
            class_ratios = {"train": train_ratio, "val": val_ratio, "test": test_size}
            split_lengths = split_lengths_from_ratios(class_ratios, num_labels)

            test_class_rule = get_single_class_rule(int(test_start_idx*num_labels), split_lengths["test"], sorted_label_set)
            traintest_dict["test"] = create_label_based_subset(dataset, labels, test_class_rule)
            split_lengths.pop("test", None)
            exclusion_rule = lambda label: not test_class_rule(label)
            sorted_label_set = sorted([x for x in sorted_label_set if exclusion_rule(x)])
            num_labels = len(sorted_label_set)

            start_idx = int((float(partition)/num_training_partitions)*num_labels)
            class_rules = get_class_rules(start_idx, split_lengths, sorted_label_set, exclusion_rule)
            for k, class_rule in class_rules.items():
                traintest_dict[k] = create_label_based_subset(dataset, labels, class_rule)

    return traintest_dict



#! /usr/bin/env python3

import collections
import errno
import glob
import os
import itertools

import numpy as np
import torch
import yaml
from easy_module_attribute_getter import utils as emag_utils
import logging
import inspect
import pytorch_metric_learning.utils.common_functions as pml_cf
import datetime
import sqlite3


def move_optimizer_to_gpu(optimizer, device):
    for state in optimizer.state.values():
        for k, v in state.items():
            if torch.is_tensor(v):
                state[k] = v.to(device)


def makedir_if_not_there(dir_name):
    try:
        os.makedirs(dir_name)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def load_yaml(fname):
    with open(fname, 'r') as f:
        loaded_yaml = yaml.safe_load(f)
    return loaded_yaml


def write_yaml(fname, input_dict, open_as):
    with open(fname, open_as) as outfile:
        yaml.dump(input_dict, outfile, default_flow_style=False, sort_keys=False)


def latest_sub_experiment_epochs(sub_experiment_dir_dict):
    latest_epochs = {}
    for sub_experiment_name, folders in sub_experiment_dir_dict.items():
        model_folder = folders[0]
        latest_epochs[sub_experiment_name] = pml_cf.latest_version(model_folder, "trunk_*.pth") or 0
    return latest_epochs


def get_all_resume_training_config_diffs(config_folder, split_scheme_name, num_training_sets):
    folder_base_name = "resume_training_config_diffs_"
    full_base_path = os.path.join(config_folder, folder_base_name)
    config_diffs = sorted(glob.glob("%s*"%full_base_path))
    if num_training_sets > 1:
        split_scheme_names = ["%s%d"%(split_scheme,i) for (split_scheme,i) in list(itertools.product(split_scheme_name, range(num_training_sets)))]
    else:
        split_scheme_names = [split_scheme_name]
    resume_training_dict = {}
    for k in config_diffs:
        latest_epochs = [int(x) for x in k.replace(full_base_path,"").split('_')]
        resume_training_dict[k] = {split_scheme:epoch for (split_scheme,epoch) in zip(split_scheme_names,latest_epochs)}
    return resume_training_dict


def save_config_files(config_folder, dict_of_yamls, resume_training, reproduce_results, latest_epochs):
    makedir_if_not_there(config_folder)
    new_dir = None
    for k, v in dict_of_yamls.items():
        k_split = k.split('/')
        config_category_name = k_split[-2] 
        if not config_category_name.startswith('config_'):
            config_category_name = os.path.splitext(k_split[-1])[0]
        fname = os.path.join(config_folder, '%s.yaml'%config_category_name)
        if not resume_training:
            if os.path.isfile(fname):
                v = emag_utils.merge_two_dicts(load_yaml(fname), v, max_merge_depth=0 if reproduce_results else float('inf'))
            write_yaml(fname, v, 'w')
        else:
            curr_yaml = load_yaml(fname)
            yaml_diff = {}
            for k2, v2 in v.items():
                if (k2 not in curr_yaml) or (v2 != curr_yaml[k2]):
                    yaml_diff[k2] = v2
            if yaml_diff != {}:
                new_dir = os.path.join(config_folder, 'resume_training_config_diffs_' + '_'.join([str(epoch) for epoch in latest_epochs]))
                makedir_if_not_there(new_dir)
                fname = os.path.join(new_dir, '%s.yaml' %config_category_name)
                write_yaml(fname, yaml_diff, 'a')

def get_last_linear(input_model, return_name=False):
    for name in ["fc", "last_linear"]:
        last_layer = getattr(input_model, name, None)
        if last_layer:
            if return_name:
                return last_layer, name
            return last_layer

def set_last_linear(input_model, set_to):
    setattr(input_model, get_last_linear(input_model, return_name=True)[1], set_to)


def check_init_arguments(input_obj, str_to_check):
    obj_stack = [input_obj]
    while len(obj_stack) > 0:
        curr_obj = obj_stack.pop()
        obj_stack += list(curr_obj.__bases__)
        if str_to_check in str(inspect.signature(curr_obj.__init__)):
            return True
    return False


def try_getting_db_count(record_keeper, table_name):
    try:
        len_of_existing_record = record_keeper.query("SELECT count(*) FROM %s"%table_name, use_global_db=False)[0]["count(*)"] 
    except sqlite3.OperationalError:
        len_of_existing_record = 0
    return len_of_existing_record


def get_datetime():
    return datetime.datetime.now()


#! /usr/bin/env python3

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
import scipy.io as sio
import os

class Cars196(Dataset):
    def __init__(self, dataset_root, transform=None):
        self.dataset_folder = os.path.join(dataset_root, "cars196")
        self.load_labels()
        self.transform = transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        path = self.img_paths[idx]
        img = Image.open(path).convert("RGB")
        label = self.labels[idx]
        if self.transform is not None:
            img = self.transform(img)
        output_dict = {"data": img, "label": label}
        return output_dict

    def load_labels(self):
        img_data = sio.loadmat(os.path.join(self.dataset_folder, "cars_annos.mat"))
        self.labels = np.array([i[0, 0] for i in img_data["annotations"]["class"][0]])
        self.img_paths = [os.path.join(self.dataset_folder, i[0]) for i in img_data["annotations"]["relative_im_path"][0]]
        self.class_names = [i[0] for i in img_data["class_names"][0]]



import setuptools
import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="powerful-benchmarker",
    version="0.9.20",
    author="Kevin Musgrave",
    author_email="tkm45@cornell.edu",
    description="A highly-configurable tool that enables thorough evaluation of deep metric learning algorithms. ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KevinMusgrave/powerful-benchmarker",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data = {'src/powerful_benchmarker': glob.glob("configs/**/*.yaml", recursive=True)},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
    install_requires=[
          'numpy',
          'scikit-learn',
          'torch',
          'torchvision',
          'easy-module-attribute-getter >= 0.9.34',
          'record-keeper >= 0.9.23',
          'tensorboard',
          'matplotlib',
          'pretrainedmodels',
          'pytorch-metric-learning >= 0.9.83',
          'pandas',
          'ax-platform',
          'faiss-gpu'
    ],
)


from .base_api_parser import BaseAPIParser

class APIUnsupervisedEmbeddingsUsingAugmentations(BaseAPIParser):
    def get_trainer_kwargs(self):
        trainer_kwargs = super().get_trainer_kwargs()
        transforms = self.get_transforms()
        trainer_kwargs["transforms"] = [transforms[k] for k in transforms.keys() if k.startswith("augmentation")]
        trainer_kwargs["sampler"] = None
        trainer_kwargs["set_min_label_to_zero"] = False
        trainer_kwargs["data_and_label_setter"] = lambda x: {"data":x[0], "label":x[1]}
        return trainer_kwargs


#! /usr/bin/env python3

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
import os

class CelebA(Dataset):
    def __init__(self, dataset_root, transform=None):
        self.dataset_folder = os.path.join(dataset_root, "celeb_a")
        self.load_labels()
        self.transform = transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        path = self.img_paths[idx]
        img = Image.open(path).convert("RGB")
        label = self.labels[idx]
        if self.transform is not None:
            img = self.transform(img)
        output_dict = {"data": img, "label": label}
        return output_dict

    def load_labels(self):
        from pandas import read_csv
        splits_info = read_csv(os.path.join(self.dataset_folder, "list_eval_partition.txt"), delim_whitespace=True, header=None).values
        self.img_paths = [os.path.join(self.dataset_folder, "img_align_celeba", x) for x in splits_info[:, 0]]
        attributes = read_csv(os.path.join(self.dataset_folder, "list_attr_celeba.txt"), delim_whitespace=True, header=1)
        self.labels = (attributes.values + 1) // 2
        self.label_names = list(attributes.columns)
        splits = splits_info[:, 1]
        all_idx = np.arange(len(splits))
        self.predefined_splits = {"train": 0, "val": 1, "test": 2}
        self.predefined_splits = {k:all_idx[splits==v] for k,v in self.predefined_splits.items()}